@@ -26,12 +26,15 @@
 
 use super::Result;
 
+#[cfg(feature = "qlog")]
+use qlog::events::h3::Http3Frame;
+
 pub const DATA_FRAME_TYPE_ID: u64 = 0x0;
 pub const HEADERS_FRAME_TYPE_ID: u64 = 0x1;
 pub const CANCEL_PUSH_FRAME_TYPE_ID: u64 = 0x3;
 pub const SETTINGS_FRAME_TYPE_ID: u64 = 0x4;
 pub const PUSH_PROMISE_FRAME_TYPE_ID: u64 = 0x5;
-pub const GOAWAY_FRAME_TYPE_ID: u64 = 0x6;
+pub const GOAWAY_FRAME_TYPE_ID: u64 = 0x7;
 pub const MAX_PUSH_FRAME_TYPE_ID: u64 = 0xD;
 pub const PRIORITY_UPDATE_FRAME_REQUEST_TYPE_ID: u64 = 0xF0700;
 pub const PRIORITY_UPDATE_FRAME_PUSH_TYPE_ID: u64 = 0xF0701;
@@ -39,12 +42,14 @@ pub const PRIORITY_UPDATE_FRAME_PUSH_TYPE_ID: u64 = 0xF0701;
 pub const SETTINGS_QPACK_MAX_TABLE_CAPACITY: u64 = 0x1;
 pub const SETTINGS_MAX_FIELD_SECTION_SIZE: u64 = 0x6;
 pub const SETTINGS_QPACK_BLOCKED_STREAMS: u64 = 0x7;
-pub const SETTINGS_H3_DATAGRAM: u64 = 0x276;
+pub const SETTINGS_ENABLE_CONNECT_PROTOCOL: u64 = 0x8;
+pub const SETTINGS_H3_DATAGRAM_00: u64 = 0x276;
+pub const SETTINGS_H3_DATAGRAM: u64 = 0x33;
 
 // Permit between 16 maximally-encoded and 128 minimally-encoded SETTINGS.
 const MAX_SETTINGS_PAYLOAD_SIZE: usize = 256;
 
-#[derive(Clone, PartialEq)]
+#[derive(Clone, PartialEq, Eq)]
 pub enum Frame {
     Data {
         payload: Vec<u8>,
@@ -62,8 +67,10 @@ pub enum Frame {
         max_field_section_size: Option<u64>,
         qpack_max_table_capacity: Option<u64>,
         qpack_blocked_streams: Option<u64>,
+        connect_protocol_enabled: Option<u64>,
         h3_datagram: Option<u64>,
         grease: Option<(u64, u64)>,
+        additional_settings: Option<Vec<(u64, u64)>>,
         raw: Option<Vec<(u64, u64)>>,
     },
 
@@ -90,7 +97,10 @@ pub enum Frame {
         priority_field_value: Vec<u8>,
     },
 
-    Unknown,
+    Unknown {
+        raw_type: u64,
+        payload: Vec<u8>,
+    },
 }
 
 impl Frame {
@@ -131,7 +141,10 @@ impl Frame {
             PRIORITY_UPDATE_FRAME_PUSH_TYPE_ID =>
                 parse_priority_update(frame_type, payload_length, &mut b)?,
 
-            _ => Frame::Unknown,
+            _ => Frame::Unknown {
+                raw_type: frame_type,
+                payload: b.get_bytes(payload_length as usize)?.to_vec(),
+            },
         };
 
         Ok(frame)
@@ -166,8 +179,10 @@ impl Frame {
                 max_field_section_size,
                 qpack_max_table_capacity,
                 qpack_blocked_streams,
+                connect_protocol_enabled,
                 h3_datagram,
                 grease,
+                additional_settings,
                 ..
             } => {
                 let mut len = 0;
@@ -187,7 +202,14 @@ impl Frame {
                     len += octets::varint_len(*val);
                 }
 
+                if let Some(val) = connect_protocol_enabled {
+                    len += octets::varint_len(SETTINGS_ENABLE_CONNECT_PROTOCOL);
+                    len += octets::varint_len(*val);
+                }
+
                 if let Some(val) = h3_datagram {
+                    len += octets::varint_len(SETTINGS_H3_DATAGRAM_00);
+                    len += octets::varint_len(*val);
                     len += octets::varint_len(SETTINGS_H3_DATAGRAM);
                     len += octets::varint_len(*val);
                 }
@@ -197,33 +219,54 @@ impl Frame {
                     len += octets::varint_len(val.1);
                 }
 
+                if let Some(vals) = additional_settings {
+                    for val in vals {
+                        len += octets::varint_len(val.0);
+                        len += octets::varint_len(val.1);
+                    }
+                }
+
                 b.put_varint(SETTINGS_FRAME_TYPE_ID)?;
                 b.put_varint(len as u64)?;
 
                 if let Some(val) = max_field_section_size {
                     b.put_varint(SETTINGS_MAX_FIELD_SECTION_SIZE)?;
-                    b.put_varint(*val as u64)?;
+                    b.put_varint(*val)?;
                 }
 
                 if let Some(val) = qpack_max_table_capacity {
                     b.put_varint(SETTINGS_QPACK_MAX_TABLE_CAPACITY)?;
-                    b.put_varint(*val as u64)?;
+                    b.put_varint(*val)?;
                 }
 
                 if let Some(val) = qpack_blocked_streams {
                     b.put_varint(SETTINGS_QPACK_BLOCKED_STREAMS)?;
-                    b.put_varint(*val as u64)?;
+                    b.put_varint(*val)?;
+                }
+
+                if let Some(val) = connect_protocol_enabled {
+                    b.put_varint(SETTINGS_ENABLE_CONNECT_PROTOCOL)?;
+                    b.put_varint(*val)?;
                 }
 
                 if let Some(val) = h3_datagram {
+                    b.put_varint(SETTINGS_H3_DATAGRAM_00)?;
+                    b.put_varint(*val)?;
                     b.put_varint(SETTINGS_H3_DATAGRAM)?;
-                    b.put_varint(*val as u64)?;
+                    b.put_varint(*val)?;
                 }
 
                 if let Some(val) = grease {
                     b.put_varint(val.0)?;
                     b.put_varint(val.1)?;
                 }
+
+                if let Some(vals) = additional_settings {
+                    for val in vals {
+                        b.put_varint(val.0)?;
+                        b.put_varint(val.1)?;
+                    }
+                }
             },
 
             Frame::PushPromise {
@@ -262,7 +305,7 @@ impl Frame {
                 b.put_varint(PRIORITY_UPDATE_FRAME_REQUEST_TYPE_ID)?;
                 b.put_varint(len as u64)?;
 
-                b.put_varint(*prioritized_element_id as u64)?;
+                b.put_varint(*prioritized_element_id)?;
                 b.put_bytes(priority_field_value)?;
             },
 
@@ -276,40 +319,177 @@ impl Frame {
                 b.put_varint(PRIORITY_UPDATE_FRAME_PUSH_TYPE_ID)?;
                 b.put_varint(len as u64)?;
 
-                b.put_varint(*prioritized_element_id as u64)?;
+                b.put_varint(*prioritized_element_id)?;
                 b.put_bytes(priority_field_value)?;
             },
 
-            Frame::Unknown => unreachable!(),
+            Frame::Unknown { raw_type, payload } => {
+                b.put_varint(*raw_type)?;
+                b.put_varint(payload.len() as u64)?;
+
+                b.put_bytes(payload.as_ref())?;
+            },
         }
 
         Ok(before - b.cap())
     }
+
+    #[cfg(feature = "qlog")]
+    pub fn to_qlog(&self) -> Http3Frame {
+        use qlog::events::RawInfo;
+
+        match self {
+            Frame::Data { .. } => Http3Frame::Data { raw: None },
+
+            // Qlog expects the `headers` to be represented as an array of
+            // name:value pairs. At this stage, we only have the qpack block, so
+            // populate the field with an empty vec.
+            Frame::Headers { .. } => Http3Frame::Headers { headers: vec![] },
+
+            Frame::CancelPush { push_id } =>
+                Http3Frame::CancelPush { push_id: *push_id },
+
+            Frame::Settings {
+                max_field_section_size,
+                qpack_max_table_capacity,
+                qpack_blocked_streams,
+                connect_protocol_enabled,
+                h3_datagram,
+                grease,
+                additional_settings,
+                ..
+            } => {
+                let mut settings = vec![];
+
+                if let Some(v) = max_field_section_size {
+                    settings.push(qlog::events::h3::Setting {
+                        name: "MAX_FIELD_SECTION_SIZE".to_string(),
+                        value: *v,
+                    });
+                }
+
+                if let Some(v) = qpack_max_table_capacity {
+                    settings.push(qlog::events::h3::Setting {
+                        name: "QPACK_MAX_TABLE_CAPACITY".to_string(),
+                        value: *v,
+                    });
+                }
+
+                if let Some(v) = qpack_blocked_streams {
+                    settings.push(qlog::events::h3::Setting {
+                        name: "QPACK_BLOCKED_STREAMS".to_string(),
+                        value: *v,
+                    });
+                }
+
+                if let Some(v) = connect_protocol_enabled {
+                    settings.push(qlog::events::h3::Setting {
+                        name: "SETTINGS_ENABLE_CONNECT_PROTOCOL".to_string(),
+                        value: *v,
+                    });
+                }
+
+                if let Some(v) = h3_datagram {
+                    settings.push(qlog::events::h3::Setting {
+                        name: "H3_DATAGRAM".to_string(),
+                        value: *v,
+                    });
+                }
+
+                if let Some((k, v)) = grease {
+                    settings.push(qlog::events::h3::Setting {
+                        name: k.to_string(),
+                        value: *v,
+                    });
+                }
+
+                if let Some(additional_settings) = additional_settings {
+                    for (k, v) in additional_settings {
+                        settings.push(qlog::events::h3::Setting {
+                            name: k.to_string(),
+                            value: *v,
+                        });
+                    }
+                }
+
+                qlog::events::h3::Http3Frame::Settings { settings }
+            },
+
+            // Qlog expects the `headers` to be represented as an array of
+            // name:value pairs. At this stage, we only have the qpack block, so
+            // populate the field with an empty vec.
+            Frame::PushPromise { push_id, .. } => Http3Frame::PushPromise {
+                push_id: *push_id,
+                headers: vec![],
+            },
+
+            Frame::GoAway { id } => Http3Frame::Goaway { id: *id },
+
+            Frame::MaxPushId { push_id } =>
+                Http3Frame::MaxPushId { push_id: *push_id },
+
+            Frame::PriorityUpdateRequest {
+                prioritized_element_id,
+                priority_field_value,
+            } => Http3Frame::PriorityUpdate {
+                target_stream_type:
+                    qlog::events::h3::H3PriorityTargetStreamType::Request,
+                prioritized_element_id: *prioritized_element_id,
+                priority_field_value: String::from_utf8_lossy(
+                    priority_field_value,
+                )
+                .into_owned(),
+            },
+
+            Frame::PriorityUpdatePush {
+                prioritized_element_id,
+                priority_field_value,
+            } => Http3Frame::PriorityUpdate {
+                target_stream_type:
+                    qlog::events::h3::H3PriorityTargetStreamType::Request,
+                prioritized_element_id: *prioritized_element_id,
+                priority_field_value: String::from_utf8_lossy(
+                    priority_field_value,
+                )
+                .into_owned(),
+            },
+
+            Frame::Unknown { raw_type, payload } => Http3Frame::Unknown {
+                frame_type_value: *raw_type,
+                raw: Some(RawInfo {
+                    data: None,
+                    payload_length: Some(payload.len() as u64),
+                    length: None,
+                }),
+            },
+        }
+    }
 }
 
 impl std::fmt::Debug for Frame {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
-            Frame::Data { payload } => {
-                write!(f, "DATA len={}", payload.len())?;
+            Frame::Data { .. } => {
+                write!(f, "DATA")?;
             },
 
-            Frame::Headers { header_block } => {
-                write!(f, "HEADERS len={}", header_block.len())?;
+            Frame::Headers { .. } => {
+                write!(f, "HEADERS")?;
             },
 
             Frame::CancelPush { push_id } => {
-                write!(f, "CANCEL_PUSH push_id={}", push_id)?;
+                write!(f, "CANCEL_PUSH push_id={push_id}")?;
             },
 
             Frame::Settings {
                 max_field_section_size,
                 qpack_max_table_capacity,
                 qpack_blocked_streams,
+                additional_settings,
                 raw,
                 ..
             } => {
-                write!(f, "SETTINGS max_field_section={:?}, qpack_max_table={:?}, qpack_blocked={:?} raw={:?}", max_field_section_size, qpack_max_table_capacity, qpack_blocked_streams, raw)?;
+                write!(f, "SETTINGS max_field_section={max_field_section_size:?}, qpack_max_table={qpack_max_table_capacity:?}, qpack_blocked={qpack_blocked_streams:?} raw={raw:?}, additional_settings={additional_settings:?}")?;
             },
 
             Frame::PushPromise {
@@ -325,11 +505,11 @@ impl std::fmt::Debug for Frame {
             },
 
             Frame::GoAway { id } => {
-                write!(f, "GOAWAY id={}", id)?;
+                write!(f, "GOAWAY id={id}")?;
             },
 
             Frame::MaxPushId { push_id } => {
-                write!(f, "MAX_PUSH_ID push_id={}", push_id)?;
+                write!(f, "MAX_PUSH_ID push_id={push_id}")?;
             },
 
             Frame::PriorityUpdateRequest {
@@ -356,8 +536,8 @@ impl std::fmt::Debug for Frame {
                 )?;
             },
 
-            Frame::Unknown => {
-                write!(f, "UNKNOWN")?;
+            Frame::Unknown { raw_type, .. } => {
+                write!(f, "UNKNOWN raw_type={raw_type}",)?;
             },
         }
 
@@ -371,8 +551,10 @@ fn parse_settings_frame(
     let mut max_field_section_size = None;
     let mut qpack_max_table_capacity = None;
     let mut qpack_blocked_streams = None;
+    let mut connect_protocol_enabled = None;
     let mut h3_datagram = None;
     let mut raw = Vec::new();
+    let mut additional_settings: Option<Vec<(u64, u64)>> = None;
 
     // Reject SETTINGS frames that are too long.
     if settings_length > MAX_SETTINGS_PAYLOAD_SIZE {
@@ -400,7 +582,15 @@ fn parse_settings_frame(
                 qpack_blocked_streams = Some(value);
             },
 
-            SETTINGS_H3_DATAGRAM => {
+            SETTINGS_ENABLE_CONNECT_PROTOCOL => {
+                if value > 1 {
+                    return Err(super::Error::SettingsError);
+                }
+
+                connect_protocol_enabled = Some(value);
+            },
+
+            SETTINGS_H3_DATAGRAM_00 | SETTINGS_H3_DATAGRAM => {
                 if value > 1 {
                     return Err(super::Error::SettingsError);
                 }
@@ -412,8 +602,12 @@ fn parse_settings_frame(
             0x0 | 0x2 | 0x3 | 0x4 | 0x5 =>
                 return Err(super::Error::SettingsError),
 
-            // Unknown Settings parameters must be ignored.
-            _ => (),
+            // Unknown Settings parameters go into additional_settings.
+            _ => {
+                let s: &mut Vec<(u64, u64)> =
+                    additional_settings.get_or_insert(vec![]);
+                s.push((identifier, value));
+            },
         }
     }
 
@@ -421,9 +615,11 @@ fn parse_settings_frame(
         max_field_section_size,
         qpack_max_table_capacity,
         qpack_blocked_streams,
+        connect_protocol_enabled,
         h3_datagram,
         grease: None,
         raw: Some(raw),
+        additional_settings,
     })
 }
 
@@ -560,6 +756,8 @@ mod tests {
             (SETTINGS_MAX_FIELD_SECTION_SIZE, 0),
             (SETTINGS_QPACK_MAX_TABLE_CAPACITY, 0),
             (SETTINGS_QPACK_BLOCKED_STREAMS, 0),
+            (SETTINGS_ENABLE_CONNECT_PROTOCOL, 0),
+            (SETTINGS_H3_DATAGRAM_00, 0),
             (SETTINGS_H3_DATAGRAM, 0),
         ];
 
@@ -567,12 +765,14 @@ mod tests {
             max_field_section_size: Some(0),
             qpack_max_table_capacity: Some(0),
             qpack_blocked_streams: Some(0),
+            connect_protocol_enabled: Some(0),
             h3_datagram: Some(0),
             grease: None,
             raw: Some(raw_settings),
+            additional_settings: None,
         };
 
-        let frame_payload_len = 9;
+        let frame_payload_len = 13;
         let frame_header_len = 2;
 
         let wire_len = {
@@ -601,15 +801,19 @@ mod tests {
             max_field_section_size: Some(0),
             qpack_max_table_capacity: Some(0),
             qpack_blocked_streams: Some(0),
+            connect_protocol_enabled: Some(0),
             h3_datagram: Some(0),
             grease: Some((33, 33)),
             raw: Default::default(),
+            additional_settings: None,
         };
 
         let raw_settings = vec![
             (SETTINGS_MAX_FIELD_SECTION_SIZE, 0),
             (SETTINGS_QPACK_MAX_TABLE_CAPACITY, 0),
             (SETTINGS_QPACK_BLOCKED_STREAMS, 0),
+            (SETTINGS_ENABLE_CONNECT_PROTOCOL, 0),
+            (SETTINGS_H3_DATAGRAM_00, 0),
             (SETTINGS_H3_DATAGRAM, 0),
             (33, 33),
         ];
@@ -620,12 +824,14 @@ mod tests {
             max_field_section_size: Some(0),
             qpack_max_table_capacity: Some(0),
             qpack_blocked_streams: Some(0),
+            connect_protocol_enabled: Some(0),
             h3_datagram: Some(0),
             grease: None,
             raw: Some(raw_settings),
+            additional_settings: Some(vec![(33, 33)]),
         };
 
-        let frame_payload_len = 11;
+        let frame_payload_len = 15;
         let frame_header_len = 2;
 
         let wire_len = {
@@ -656,9 +862,11 @@ mod tests {
             max_field_section_size: Some(1024),
             qpack_max_table_capacity: None,
             qpack_blocked_streams: None,
+            connect_protocol_enabled: None,
             h3_datagram: None,
             grease: None,
             raw: Some(raw_settings),
+            additional_settings: None,
         };
 
         let frame_payload_len = 3;
@@ -683,21 +891,99 @@ mod tests {
     }
 
     #[test]
-    fn settings_h3_dgram_only() {
+    fn settings_h3_connect_protocol_enabled() {
         let mut d = [42; 128];
 
-        let raw_settings = vec![(SETTINGS_H3_DATAGRAM, 1)];
+        let raw_settings = vec![(SETTINGS_ENABLE_CONNECT_PROTOCOL, 1)];
 
         let frame = Frame::Settings {
             max_field_section_size: None,
             qpack_max_table_capacity: None,
             qpack_blocked_streams: None,
+            connect_protocol_enabled: Some(1),
+            h3_datagram: None,
+            grease: None,
+            raw: Some(raw_settings),
+            additional_settings: None,
+        };
+
+        let frame_payload_len = 2;
+        let frame_header_len = 2;
+
+        let wire_len = {
+            let mut b = octets::OctetsMut::with_slice(&mut d);
+            frame.to_bytes(&mut b).unwrap()
+        };
+
+        assert_eq!(wire_len, frame_header_len + frame_payload_len);
+
+        assert_eq!(
+            Frame::from_bytes(
+                SETTINGS_FRAME_TYPE_ID,
+                frame_payload_len as u64,
+                &d[frame_header_len..]
+            )
+            .unwrap(),
+            frame
+        );
+    }
+
+    #[test]
+    fn settings_h3_connect_protocol_enabled_bad() {
+        let mut d = [42; 128];
+
+        let raw_settings = vec![(SETTINGS_ENABLE_CONNECT_PROTOCOL, 9)];
+
+        let frame = Frame::Settings {
+            max_field_section_size: None,
+            qpack_max_table_capacity: None,
+            qpack_blocked_streams: None,
+            connect_protocol_enabled: Some(9),
+            h3_datagram: None,
+            grease: None,
+            raw: Some(raw_settings),
+            additional_settings: None,
+        };
+
+        let frame_payload_len = 2;
+        let frame_header_len = 2;
+
+        let wire_len = {
+            let mut b = octets::OctetsMut::with_slice(&mut d);
+            frame.to_bytes(&mut b).unwrap()
+        };
+
+        assert_eq!(wire_len, frame_header_len + frame_payload_len);
+
+        assert_eq!(
+            Frame::from_bytes(
+                SETTINGS_FRAME_TYPE_ID,
+                frame_payload_len as u64,
+                &d[frame_header_len..]
+            ),
+            Err(crate::h3::Error::SettingsError)
+        );
+    }
+
+    #[test]
+    fn settings_h3_dgram_only() {
+        let mut d = [42; 128];
+
+        let raw_settings =
+            vec![(SETTINGS_H3_DATAGRAM_00, 1), (SETTINGS_H3_DATAGRAM, 1)];
+
+        let frame = Frame::Settings {
+            max_field_section_size: None,
+            qpack_max_table_capacity: None,
+            qpack_blocked_streams: None,
+            connect_protocol_enabled: None,
             h3_datagram: Some(1),
             grease: None,
             raw: Some(raw_settings),
+            additional_settings: None,
         };
 
-        let frame_payload_len = 3;
+        let frame_payload_len = 5;
         let frame_header_len = 2;
 
         let wire_len = {
@@ -726,12 +1012,14 @@ mod tests {
             max_field_section_size: None,
             qpack_max_table_capacity: None,
             qpack_blocked_streams: None,
+            connect_protocol_enabled: None,
             h3_datagram: Some(5),
             grease: None,
             raw: Default::default(),
+            additional_settings: None,
         };
 
-        let frame_payload_len = 3;
+        let frame_payload_len = 5;
         let frame_header_len = 2;
 
         let wire_len = {
@@ -764,9 +1052,11 @@ mod tests {
             max_field_section_size: None,
             qpack_max_table_capacity: Some(0),
             qpack_blocked_streams: Some(0),
+            connect_protocol_enabled: None,
             h3_datagram: None,
             grease: None,
             raw: Some(raw_settings),
+            additional_settings: None,
         };
 
         let frame_payload_len = 4;
@@ -1036,6 +1326,12 @@ mod tests {
     fn unknown_type() {
         let d = [42; 12];
 
-        assert_eq!(Frame::from_bytes(255, 12345, &d[..]), Ok(Frame::Unknown));
+        assert_eq!(
+            Frame::from_bytes(255, 12, &d[..]),
+            Ok(Frame::Unknown {
+                raw_type: 255,
+                payload: vec![42; 12]
+            })
+        );
     }
 }
