@@ -357,6 +357,7 @@ mod tests {
     use crate::recovery::congestion::test_sender::TestSender;
     use crate::recovery::HandshakeStatus;
     use crate::recovery::RecoveryOps;
+    use crate::OnAckReceivedOutcome;
 
     use smallvec::smallvec;
 
@@ -375,7 +376,7 @@ mod tests {
             r.cwnd(),
             r.max_datagram_size * cfg.initial_congestion_window_packets
         );
-        assert_eq!(r.bytes_in_flight, 0);
+        assert_eq!(r.bytes_in_flight(), 0);
 
         assert_eq!(r.congestion.bbr_state.state, BBRStateMachine::Startup);
     }
@@ -404,10 +405,13 @@ mod tests {
         assert_eq!(sender.congestion_window, cwnd_prev + mss * 5);
         assert_eq!(sender.bytes_in_flight, 0);
         assert_eq!(
-            sender.delivery_rate(),
+            sender.delivery_rate().to_bytes_per_second(),
             ((mss * 5) as f64 / rtt.as_secs_f64()) as u64
         );
-        assert_eq!(sender.bbr_state.btlbw, sender.delivery_rate());
+        assert_eq!(
+            sender.bbr_state.btlbw,
+            sender.delivery_rate().to_bytes_per_second()
+        );
     }
 
     #[test]
@@ -437,7 +441,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -464,15 +468,22 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (2, 2 * mss, mss),
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 2,
+                lost_bytes: 2 * mss,
+                acked_bytes: mss,
+                spurious_losses: 0,
+            },
         );
 
         // Sent: 0, 1, 2, 3, 4, Acked 4.
         assert_eq!(r.cwnd(), mss * 4);
         // Stil in flight: 2, 3.
-        assert_eq!(r.bytes_in_flight, mss * 2);
+        assert_eq!(r.bytes_in_flight(), mss * 2);
     }
 
     #[test]
@@ -504,7 +515,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -531,9 +542,16 @@ mod tests {
                     packet::Epoch::Application,
                     HandshakeStatus::default(),
                     now,
+                    None,
                     "",
-                ),
-                (0, 0, mss),
+                )
+                .unwrap(),
+                OnAckReceivedOutcome {
+                    lost_packets: 0,
+                    lost_bytes: 0,
+                    acked_bytes: mss,
+                    spurious_losses: 0,
+                },
             );
         }
 
@@ -554,7 +572,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -583,9 +601,16 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, mss),
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: mss,
+                spurious_losses: 0,
+            },
         );
 
         // Now we are in Drain state.
@@ -623,7 +648,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -647,9 +672,16 @@ mod tests {
                     packet::Epoch::Application,
                     HandshakeStatus::default(),
                     now,
+                    None,
                     "",
-                ),
-                (0, 0, mss),
+                )
+                .unwrap(),
+                OnAckReceivedOutcome {
+                    lost_packets: 0,
+                    lost_bytes: 0,
+                    acked_bytes: mss,
+                    spurious_losses: 0,
+                },
             );
         }
 
@@ -692,7 +724,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -718,9 +750,16 @@ mod tests {
                     packet::Epoch::Application,
                     HandshakeStatus::default(),
                     now,
+                    None,
                     "",
-                ),
-                (0, 0, mss),
+                )
+                .unwrap(),
+                OnAckReceivedOutcome {
+                    lost_packets: 0,
+                    lost_bytes: 0,
+                    acked_bytes: mss,
+                    spurious_losses: 0,
+                },
             );
         }
 
@@ -746,7 +785,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -774,9 +813,16 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, mss),
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: mss,
+                spurious_losses: 0,
+            },
         );
 
         assert_eq!(r.congestion.bbr_state.state, BBRStateMachine::ProbeRTT);
