@@ -1,6 +1,3 @@
-// Copyright (C) 2022, Cloudflare, Inc.
-// All rights reserved.
-//
 // Redistribution and use in source and binary forms, with or without
 // modification, are permitted provided that the following conditions are
 // met:
@@ -669,6 +666,7 @@ mod tests {
     use crate::recovery::HandshakeStatus;
     use crate::recovery::RecoveryOps;
     use crate::CongestionControlAlgorithm;
+    use crate::OnAckReceivedOutcome;
 
     #[test]
     fn bbr_init() {
@@ -684,7 +682,7 @@ mod tests {
             r.cwnd(),
             r.max_datagram_size * r.congestion.initial_congestion_window_packets
         );
-        assert_eq!(r.bytes_in_flight, 0);
+        assert_eq!(r.bytes_in_flight(), 0);
 
         assert_eq!(r.congestion.bbr2_state.state, BBR2StateMachine::Startup);
     }
@@ -716,7 +714,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -742,19 +740,29 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, 5 * mss)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: mss * 5,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.congestion.bbr2_state.state, BBR2StateMachine::Startup);
         assert_eq!(r.cwnd(), cwnd_prev + mss * 5);
-        assert_eq!(r.bytes_in_flight, 0);
+        assert_eq!(r.bytes_in_flight(), 0);
         assert_eq!(
-            r.delivery_rate(),
+            r.delivery_rate().to_bytes_per_second(),
             ((mss * 5) as f64 / rtt.as_secs_f64()) as u64
         );
-        assert_eq!(r.congestion.bbr2_state.full_bw, r.delivery_rate());
+        assert_eq!(
+            r.congestion.bbr2_state.full_bw,
+            r.delivery_rate().to_bytes_per_second()
+        );
     }
 
     #[test]
@@ -784,7 +792,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -811,15 +819,22 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (2, 2 * mss, mss)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 2,
+                lost_bytes: 2 * mss,
+                acked_bytes: mss,
+                spurious_losses: 0,
+            }
         );
 
         assert!(r.congestion.bbr2_state.in_recovery);
 
         // Still in flight: 2, 3.
-        assert_eq!(r.bytes_in_flight, mss * 2);
+        assert_eq!(r.bytes_in_flight(), mss * 2);
 
         assert_eq!(r.congestion.bbr2_state.newly_acked_bytes, mss);
 
@@ -855,7 +870,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -882,9 +897,16 @@ mod tests {
                     packet::Epoch::Application,
                     HandshakeStatus::default(),
                     now,
+                    None,
                     "",
-                ),
-                (0, 0, mss)
+                )
+                .unwrap(),
+                OnAckReceivedOutcome {
+                    lost_packets: 0,
+                    lost_bytes: 0,
+                    acked_bytes: mss,
+                    spurious_losses: 0,
+                }
             );
         }
 
@@ -906,7 +928,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -936,9 +958,16 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, mss)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: mss,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.congestion.bbr2_state.state, BBR2StateMachine::Drain);
@@ -977,7 +1006,7 @@ mod tests {
                 tx_in_flight: 0,
                 lost: 0,
                 has_data: false,
-                pmtud: false,
+                is_pmtud_probe: false,
             };
 
             r.on_packet_sent(
@@ -1003,9 +1032,16 @@ mod tests {
                     packet::Epoch::Application,
                     HandshakeStatus::default(),
                     now,
+                    None,
                     "",
-                ),
-                (0, 0, mss)
+                )
+                .unwrap(),
+                OnAckReceivedOutcome {
+                    lost_packets: 0,
+                    lost_bytes: 0,
+                    acked_bytes: mss,
+                    spurious_losses: 0,
+                }
             );
         }
 
@@ -1034,7 +1070,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1062,9 +1098,16 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, mss)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: mss,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.congestion.bbr2_state.state, BBR2StateMachine::ProbeRTT);
