@@ -34,26 +34,31 @@ use std::collections::VecDeque;
 use super::RecoveryConfig;
 use super::Sent;
 
-use crate::recovery::HandshakeStatus;
-use crate::recovery::RecoveryOps;
-
 use crate::packet::Epoch;
 use crate::ranges::RangeSet;
+use crate::recovery::Bandwidth;
+use crate::recovery::HandshakeStatus;
+use crate::recovery::OnLossDetectionTimeoutOutcome;
+use crate::recovery::RecoveryOps;
+use crate::recovery::StartupExit;
+use crate::Error;
+use crate::Result;
 
 #[cfg(feature = "qlog")]
 use crate::recovery::QlogMetrics;
 
 use crate::frame;
 use crate::packet;
-use crate::ranges;
 
 #[cfg(feature = "qlog")]
 use qlog::events::EventData;
 
 use super::pacer;
 use super::Congestion;
+use crate::recovery::bytes_in_flight::BytesInFlight;
 use crate::recovery::rtt::RttStats;
 use crate::recovery::LossDetectionTimer;
+use crate::recovery::OnAckReceivedOutcome;
 use crate::recovery::ReleaseDecision;
 use crate::recovery::ReleaseTime;
 use crate::recovery::GRANULARITY;
@@ -85,6 +90,10 @@ struct RecoveryEpoch {
 
     acked_frames: Vec<frame::Frame>,
     lost_frames: Vec<frame::Frame>,
+
+    /// The largest packet number sent in the packet number space so far.
+    #[cfg(test)]
+    test_largest_sent_pkt_num_on_path: Option<u64>,
 }
 
 struct AckedDetectionResult {
@@ -103,10 +112,12 @@ struct LossDetectionResult {
 }
 
 impl RecoveryEpoch {
+    // `peer_sent_ack_ranges` should not be used without validation.
     fn detect_and_remove_acked_packets(
-        &mut self, now: Instant, acked: &RangeSet, newly_acked: &mut Vec<Acked>,
-        rtt_stats: &RttStats, trace_id: &str,
-    ) -> AckedDetectionResult {
+        &mut self, now: Instant, peer_sent_ack_ranges: &RangeSet,
+        newly_acked: &mut Vec<Acked>, rtt_stats: &RttStats, skip_pn: Option<u64>,
+        trace_id: &str,
+    ) -> Result<AckedDetectionResult> {
         newly_acked.clear();
 
         let mut acked_bytes = 0;
@@ -115,27 +126,41 @@ impl RecoveryEpoch {
         let mut has_ack_eliciting = false;
         let mut has_in_flight_spurious_loss = false;
 
-        let largest_acked = self.largest_acked_packet.unwrap();
+        let largest_ack_received = peer_sent_ack_ranges
+            .last()
+            .expect("ACK frames should always have at least one ack range");
+        let largest_acked = self
+            .largest_acked_packet
+            .unwrap_or(0)
+            .max(largest_ack_received);
+
+        for peer_sent_range in peer_sent_ack_ranges.iter() {
+            if skip_pn.is_some_and(|skip_pn| peer_sent_range.contains(&skip_pn)) {
+                // https://www.rfc-editor.org/rfc/rfc9000#section-13.1
+                // An endpoint SHOULD treat receipt of an acknowledgment
+                // for a packet it did not send as
+                // a connection error of type PROTOCOL_VIOLATION
+                return Err(Error::OptimisticAckDetected);
+            }
 
-        for ack in acked.iter() {
             // Because packets always have incrementing numbers, they are always
             // in sorted order.
             let start = if self
                 .sent_packets
                 .front()
-                .filter(|e| e.pkt_num >= ack.start)
+                .filter(|e| e.pkt_num >= peer_sent_range.start)
                 .is_some()
             {
                 // Usually it will be the first packet.
                 0
             } else {
                 self.sent_packets
-                    .binary_search_by_key(&ack.start, |p| p.pkt_num)
+                    .binary_search_by_key(&peer_sent_range.start, |p| p.pkt_num)
                     .unwrap_or_else(|e| e)
             };
 
             for unacked in self.sent_packets.range_mut(start..) {
-                if unacked.pkt_num >= ack.end {
+                if unacked.pkt_num >= peer_sent_range.end {
                     break;
                 }
 
@@ -182,13 +207,13 @@ impl RecoveryEpoch {
 
         self.drain_acked_and_lost_packets(now - rtt_stats.rtt());
 
-        AckedDetectionResult {
+        Ok(AckedDetectionResult {
             acked_bytes,
             spurious_losses,
             spurious_pkt_thresh,
             has_ack_eliciting,
             has_in_flight_spurious_loss,
-        }
+        })
     }
 
     fn detect_lost_packets(
@@ -226,7 +251,7 @@ impl RecoveryEpoch {
 
                 unacked.time_lost = Some(now);
 
-                if unacked.pmtud {
+                if unacked.is_pmtud_probe {
                     pmtud_lost_bytes += unacked.size;
                     self.in_flight_count -= 1;
 
@@ -313,7 +338,7 @@ pub struct LegacyRecovery {
 
     time_thresh: f64,
 
-    pub bytes_in_flight: usize,
+    bytes_in_flight: BytesInFlight,
 
     bytes_sent: usize,
 
@@ -350,7 +375,7 @@ impl LegacyRecovery {
 
             time_thresh: INITIAL_TIME_THRESHOLD,
 
-            bytes_in_flight: 0,
+            bytes_in_flight: Default::default(),
 
             bytes_sent: 0,
 
@@ -396,7 +421,7 @@ impl LegacyRecovery {
         let mut duration = self.pto() * 2_u32.pow(self.pto_count);
 
         // Arm PTO from now when there are no inflight packets.
-        if self.bytes_in_flight == 0 {
+        if self.bytes_in_flight.is_zero() {
             if handshake_status.has_handshake_keys {
                 return (Some(now + duration), packet::Epoch::Handshake);
             } else {
@@ -453,7 +478,9 @@ impl LegacyRecovery {
             return;
         }
 
-        if self.bytes_in_flight == 0 && handshake_status.peer_verified_address {
+        if self.bytes_in_flight.is_zero() &&
+            handshake_status.peer_verified_address
+        {
             self.loss_timer.clear();
             return;
         }
@@ -486,16 +513,18 @@ impl LegacyRecovery {
 
             (self.congestion.cc_ops.congestion_event)(
                 &mut self.congestion,
-                self.bytes_in_flight,
+                self.bytes_in_flight.get(),
                 loss.lost_bytes,
                 &pkt,
                 now,
             );
 
-            self.bytes_in_flight -= loss.lost_bytes;
+            self.bytes_in_flight
+                .saturating_subtract(loss.lost_bytes, now);
         };
 
-        self.bytes_in_flight -= loss.pmtud_lost_bytes;
+        self.bytes_in_flight
+            .saturating_subtract(loss.pmtud_lost_bytes, now);
 
         self.epochs[epoch]
             .drain_acked_and_lost_packets(now - self.rtt_stats.rtt());
@@ -564,7 +593,7 @@ impl RecoveryOps for LegacyRecovery {
         }
 
         self.congestion.on_packet_sent(
-            self.bytes_in_flight,
+            self.bytes_in_flight.get(),
             sent_bytes,
             now,
             &mut pkt,
@@ -575,38 +604,37 @@ impl RecoveryOps for LegacyRecovery {
 
         if in_flight {
             self.epochs[epoch].in_flight_count += 1;
-            self.bytes_in_flight += sent_bytes;
+            self.bytes_in_flight.add(sent_bytes, now);
 
             self.set_loss_detection_timer(handshake_status, now);
         }
 
         self.bytes_sent += sent_bytes;
 
+        #[cfg(test)]
+        {
+            self.epochs[epoch].test_largest_sent_pkt_num_on_path = self.epochs
+                [epoch]
+                .test_largest_sent_pkt_num_on_path
+                .max(Some(pkt.pkt_num));
+        }
+
         self.epochs[epoch].sent_packets.push_back(pkt);
 
         trace!("{} {:?}", trace_id, self);
     }
 
-    fn get_packet_send_time(&self) -> Instant {
+    fn get_packet_send_time(&self, _now: Instant) -> Instant {
+        // TODO .max(now)
         self.congestion.get_packet_send_time()
     }
 
-    #[allow(clippy::too_many_arguments)]
+    // `peer_sent_ack_ranges` should not be used without validation.
     fn on_ack_received(
-        &mut self, ranges: &ranges::RangeSet, ack_delay: u64,
+        &mut self, peer_sent_ack_ranges: &RangeSet, ack_delay: u64,
         epoch: packet::Epoch, handshake_status: HandshakeStatus, now: Instant,
-        trace_id: &str,
-    ) -> (usize, usize, usize) {
-        let largest_acked = ranges.last().unwrap();
-
-        // Update the largest acked packet.
-        let largest_acked = self.epochs[epoch]
-            .largest_acked_packet
-            .unwrap_or(0)
-            .max(largest_acked);
-
-        self.epochs[epoch].largest_acked_packet = Some(largest_acked);
-
+        skip_pn: Option<u64>, trace_id: &str,
+    ) -> Result<OnAckReceivedOutcome> {
         let AckedDetectionResult {
             acked_bytes,
             spurious_losses,
@@ -615,11 +643,12 @@ impl RecoveryOps for LegacyRecovery {
             has_in_flight_spurious_loss,
         } = self.epochs[epoch].detect_and_remove_acked_packets(
             now,
-            ranges,
+            peer_sent_ack_ranges,
             &mut self.newly_acked,
             &self.rtt_stats,
+            skip_pn,
             trace_id,
-        );
+        )?;
 
         self.lost_spurious_count += spurious_losses;
         if let Some(thresh) = spurious_pkt_thresh {
@@ -633,13 +662,23 @@ impl RecoveryOps for LegacyRecovery {
         }
 
         if self.newly_acked.is_empty() {
-            return (0, 0, 0);
+            return Ok(OnAckReceivedOutcome::default());
         }
 
-        // Check if largest packet is newly acked.
         let largest_newly_acked = self.newly_acked.last().unwrap();
 
-        if largest_newly_acked.pkt_num == largest_acked && has_ack_eliciting {
+        // Update `largest_acked_packet` based on the validated `newly_acked`
+        // value.
+        let largest_acked_pkt_num = self.epochs[epoch]
+            .largest_acked_packet
+            .unwrap_or(0)
+            .max(largest_newly_acked.pkt_num);
+        self.epochs[epoch].largest_acked_packet = Some(largest_acked_pkt_num);
+
+        // Check if largest packet is newly acked.
+        if largest_newly_acked.pkt_num == largest_acked_pkt_num &&
+            has_ack_eliciting
+        {
             let latest_rtt = now - largest_newly_acked.time_sent;
             self.rtt_stats.update_rtt(
                 latest_rtt,
@@ -651,16 +690,17 @@ impl RecoveryOps for LegacyRecovery {
 
         // Detect and mark lost packets without removing them from the sent
         // packets list.
-        let loss = self.detect_lost_packets(epoch, now, trace_id);
+        let (lost_packets, lost_bytes) =
+            self.detect_lost_packets(epoch, now, trace_id);
 
         self.congestion.on_packets_acked(
-            self.bytes_in_flight,
+            self.bytes_in_flight.get(),
             &mut self.newly_acked,
             &self.rtt_stats,
             now,
         );
 
-        self.bytes_in_flight -= acked_bytes;
+        self.bytes_in_flight.saturating_subtract(acked_bytes, now);
 
         self.pto_count = 0;
 
@@ -669,26 +709,35 @@ impl RecoveryOps for LegacyRecovery {
         self.epochs[epoch]
             .drain_acked_and_lost_packets(now - self.rtt_stats.rtt());
 
-        (loss.0, loss.1, acked_bytes)
+        Ok(OnAckReceivedOutcome {
+            lost_packets,
+            lost_bytes,
+            acked_bytes,
+            spurious_losses,
+        })
     }
 
     fn on_loss_detection_timeout(
         &mut self, handshake_status: HandshakeStatus, now: Instant,
         trace_id: &str,
-    ) -> (usize, usize) {
+    ) -> OnLossDetectionTimeoutOutcome {
         let (earliest_loss_time, epoch) = self.loss_time_and_space();
 
         if earliest_loss_time.is_some() {
             // Time threshold loss detection.
-            let loss = self.detect_lost_packets(epoch, now, trace_id);
+            let (lost_packets, lost_bytes) =
+                self.detect_lost_packets(epoch, now, trace_id);
 
             self.set_loss_detection_timer(handshake_status, now);
 
             trace!("{} {:?}", trace_id, self);
-            return loss;
+            return OnLossDetectionTimeoutOutcome {
+                lost_packets,
+                lost_bytes,
+            };
         }
 
-        let epoch = if self.bytes_in_flight > 0 {
+        let epoch = if self.bytes_in_flight.get() > 0 {
             // Send new data if available, else retransmit old data. If neither
             // is available, send a single PING frame.
             let (_, e) = self.pto_time_and_space(handshake_status, now);
@@ -736,7 +785,10 @@ impl RecoveryOps for LegacyRecovery {
 
         trace!("{} {:?}", trace_id, self);
 
-        (0, 0)
+        OnLossDetectionTimeoutOutcome {
+            lost_packets: 0,
+            lost_bytes: 0,
+        }
     }
 
     fn on_pkt_num_space_discarded(
@@ -753,7 +805,7 @@ impl RecoveryOps for LegacyRecovery {
             })
             .fold(0, |acc, p| acc + p.size);
 
-        self.bytes_in_flight -= unacked_bytes;
+        self.bytes_in_flight.saturating_subtract(unacked_bytes, now);
 
         epoch.sent_packets.clear();
         epoch.lost_frames.clear();
@@ -789,7 +841,7 @@ impl RecoveryOps for LegacyRecovery {
         }
 
         // Open more space (snd_cnt) for PRR when allowed.
-        self.cwnd().saturating_sub(self.bytes_in_flight) +
+        self.cwnd().saturating_sub(self.bytes_in_flight.get()) +
             self.congestion.prr.snd_cnt
     }
 
@@ -801,6 +853,10 @@ impl RecoveryOps for LegacyRecovery {
         self.rtt_stats.min_rtt()
     }
 
+    fn max_rtt(&self) -> Option<Duration> {
+        self.rtt_stats.max_rtt()
+    }
+
     fn rttvar(&self) -> Duration {
         self.rtt_stats.rttvar
     }
@@ -809,10 +865,16 @@ impl RecoveryOps for LegacyRecovery {
         self.rtt() + cmp::max(self.rtt_stats.rttvar * 4, GRANULARITY)
     }
 
-    fn delivery_rate(&self) -> u64 {
+    /// The most recent data delivery rate estimate.
+    fn delivery_rate(&self) -> Bandwidth {
         self.congestion.delivery_rate()
     }
 
+    /// Statistics from when a CCA first exited the startup phase.
+    fn startup_exit(&self) -> Option<StartupExit> {
+        self.congestion.ssthresh.startup_exit()
+    }
+
     fn max_datagram_size(&self) -> usize {
         self.max_datagram_size
     }
@@ -857,7 +919,11 @@ impl RecoveryOps for LegacyRecovery {
 
     #[cfg(test)]
     fn bytes_in_flight(&self) -> usize {
-        self.bytes_in_flight
+        self.bytes_in_flight.get()
+    }
+
+    fn bytes_in_flight_duration(&self) -> Duration {
+        self.bytes_in_flight.get_duration()
     }
 
     #[cfg(test)]
@@ -887,24 +953,32 @@ impl RecoveryOps for LegacyRecovery {
         self.detect_lost_packets(epoch, now, "")
     }
 
+    // FIXME only used by gcongestion
     fn on_app_limited(&mut self) {
         // Not implemented for legacy recovery, update_app_limited and
         // delivery_rate_update_app_limited used instead.
     }
 
+    #[cfg(test)]
+    fn largest_sent_pkt_num_on_path(&self, epoch: packet::Epoch) -> Option<u64> {
+        self.epochs[epoch].test_largest_sent_pkt_num_on_path
+    }
+
     #[cfg(test)]
     fn app_limited(&self) -> bool {
         self.congestion.app_limited
     }
 
     fn update_app_limited(&mut self, v: bool) {
-        self.congestion.app_limited = v;
+        self.congestion.update_app_limited(v);
     }
 
+    // FIXME only used by congestion
     fn delivery_rate_update_app_limited(&mut self, v: bool) {
         self.congestion.delivery_rate.update_app_limited(v);
     }
 
+    // FIXME only used by congestion
     fn update_max_ack_delay(&mut self, max_ack_delay: Duration) {
         self.rtt_stats.max_ack_delay = max_ack_delay;
     }
@@ -917,8 +991,8 @@ impl RecoveryOps for LegacyRecovery {
             latest_rtt: self.rtt_stats.latest_rtt,
             rttvar: self.rtt_stats.rttvar,
             cwnd: self.cwnd() as u64,
-            bytes_in_flight: self.bytes_in_flight as u64,
-            ssthresh: Some(self.congestion.ssthresh as u64),
+            bytes_in_flight: self.bytes_in_flight.get() as u64,
+            ssthresh: Some(self.congestion.ssthresh.get() as u64),
             pacing_rate: self.congestion.pacer.rate(),
         };
 
@@ -967,8 +1041,8 @@ impl std::fmt::Debug for LegacyRecovery {
         write!(f, "min_rtt={:?} ", *self.rtt_stats.min_rtt)?;
         write!(f, "rttvar={:?} ", self.rtt_stats.rttvar)?;
         write!(f, "cwnd={} ", self.cwnd())?;
-        write!(f, "ssthresh={} ", self.congestion.ssthresh)?;
-        write!(f, "bytes_in_flight={} ", self.bytes_in_flight)?;
+        write!(f, "ssthresh={} ", self.congestion.ssthresh.get())?;
+        write!(f, "bytes_in_flight={} ", self.bytes_in_flight.get())?;
         write!(f, "app_limited={} ", self.congestion.app_limited)?;
         write!(
             f,
