@@ -40,10 +40,12 @@ use std::time::Instant;
 
 use network_model::BBRv2NetworkModel;
 
+use crate::recovery::gcongestion::Bandwidth;
+use crate::recovery::RecoveryStats;
+
 use self::mode::Mode;
 use self::mode::ModeImpl;
 
-use super::bandwidth::Bandwidth;
 use super::bbr::SendTimeState;
 use super::Acked;
 use super::BbrBwLoReductionStrategy;
@@ -67,6 +69,8 @@ struct Params {
     /// trips.
     full_bw_threshold: f32,
 
+    /// The number of rounds to stay in  STARTUP before exiting due to
+    /// bandwidth plateau.
     startup_full_bw_rounds: usize,
 
     /// Number of rounds to stay in STARTUP when there's a sufficient queue that
@@ -157,9 +161,18 @@ struct Params {
     decrease_startup_pacing_at_end_of_round: bool,
 
     /// Avoid Overestimation in Bandwidth Sampler with ack aggregation
-    overestimate_avoidance: bool,
+    enable_overestimate_avoidance: bool,
+
+    /// If true, apply the fix to A0 point selection logic so the
+    /// implementation is consistent with the behavior of the
+    /// google/quiche implementation.
+    choose_a0_point_fix: bool,
 
     bw_lo_mode: BwLoMode,
+
+    /// Determines whether app limited rounds with no bandwidth growth count
+    /// towards the rounds threshold to exit startup.
+    ignore_app_limited_for_no_bandwidth_growth: bool,
 }
 
 impl Params {
@@ -179,6 +192,8 @@ impl Params {
         apply_override!(drain_cwnd_gain);
         apply_override!(drain_pacing_gain);
         apply_override!(enable_reno_coexistence);
+        apply_override!(enable_overestimate_avoidance);
+        apply_override!(choose_a0_point_fix);
         apply_override!(probe_bw_probe_up_pacing_gain);
         apply_override!(probe_bw_probe_down_pacing_gain);
         apply_override!(probe_bw_cwnd_gain);
@@ -186,6 +201,7 @@ impl Params {
         apply_override!(loss_threshold);
         apply_override!(use_bytes_delivered_for_inflight_hi);
         apply_override!(decrease_startup_pacing_at_end_of_round);
+        apply_override!(ignore_app_limited_for_no_bandwidth_growth);
 
         if let Some(custom_value) = custom_bbr_settings.bw_lo_reduction_strategy {
             self.bw_lo_mode = custom_value.into();
@@ -260,9 +276,13 @@ const DEFAULT_PARAMS: Params = Params {
 
     decrease_startup_pacing_at_end_of_round: true,
 
-    overestimate_avoidance: true,
+    enable_overestimate_avoidance: true,
+
+    choose_a0_point_fix: false,
 
     bw_lo_mode: BwLoMode::InflightReduction,
+
+    ignore_app_limited_for_no_bandwidth_growth: false,
 };
 
 #[derive(Debug, PartialEq)]
@@ -536,6 +556,7 @@ impl CongestionControl for BBRv2 {
         &mut self, _rtt_updated: bool, prior_in_flight: usize,
         _bytes_in_flight: usize, event_time: Instant, acked_packets: &[Acked],
         lost_packets: &[Lost], least_unacked: u64, _rtt_stats: &RttStats,
+        recovery_stats: &mut RecoveryStats,
     ) {
         let mut congestion_event = BBRv2CongestionEvent::new(
             event_time,
@@ -562,6 +583,8 @@ impl CongestionControl for BBRv2 {
                 &mut congestion_event,
                 self.target_bytes_inflight(),
                 &self.params,
+                recovery_stats,
+                self.get_congestion_window(),
             )
         {
             mode_changes_allowed -= 1;
