@@ -32,9 +32,9 @@ use std::ops::Add;
 use std::time::Duration;
 use std::time::Instant;
 
-use crate::recovery::gcongestion::bandwidth::Bandwidth;
 use crate::recovery::gcongestion::bbr::BandwidthSampler;
 use crate::recovery::gcongestion::bbr2::Params;
+use crate::recovery::gcongestion::Bandwidth;
 use crate::recovery::gcongestion::Lost;
 use crate::recovery::rtt::RttStats;
 use crate::recovery::rtt::INITIAL_RTT;
@@ -182,8 +182,12 @@ pub(super) struct BBRv2NetworkModel {
     full_bandwidth_baseline: Bandwidth,
     rounds_without_bandwidth_growth: usize,
 
-    // Used by STARTUP and PROBE_UP to decide when to exit.
+    /// Used by STARTUP and PROBE_UP to decide when to exit.
     rounds_with_queueing: usize,
+
+    /// Determines whether app limited rounds with no bandwidth growth count
+    /// towards the rounds threshold to exit startup.
+    ignore_app_limited_for_no_bandwidth_growth: bool,
 }
 
 impl BBRv2NetworkModel {
@@ -193,7 +197,8 @@ impl BBRv2NetworkModel {
             inflight_hi_limited_in_round: false,
             bandwidth_sampler: BandwidthSampler::new(
                 params.initial_max_ack_height_filter_window,
-                params.overestimate_avoidance,
+                params.enable_overestimate_avoidance,
+                params.choose_a0_point_fix,
             ),
             round_trip_counter: RoundTripCounter {
                 round_trip_count: 0,
@@ -224,6 +229,9 @@ impl BBRv2NetworkModel {
             full_bandwidth_baseline: Bandwidth::zero(),
             rounds_without_bandwidth_growth: 0,
             rounds_with_queueing: 0,
+
+            ignore_app_limited_for_no_bandwidth_growth: params
+                .ignore_app_limited_for_no_bandwidth_growth,
         }
     }
 
@@ -611,7 +619,12 @@ impl BBRv2NetworkModel {
             return true;
         }
 
-        self.rounds_without_bandwidth_growth += 1;
+        let ignore_round = self.ignore_app_limited_for_no_bandwidth_growth &&
+            congestion_event.last_packet_send_state.is_app_limited;
+
+        if !ignore_round {
+            self.rounds_without_bandwidth_growth += 1;
+        }
 
         // full_bandwidth_reached is only set to true when not app-limited
         if self.rounds_without_bandwidth_growth >= params.startup_full_bw_rounds &&
