@@ -1,4 +1,5 @@
 use crate::packet;
+use crate::recovery::OnLossDetectionTimeoutOutcome;
 
 use std::collections::VecDeque;
 use std::time::Duration;
@@ -14,23 +15,29 @@ use crate::recovery::QlogMetrics;
 
 use crate::frame;
 
+use crate::recovery::bytes_in_flight::BytesInFlight;
+use crate::recovery::gcongestion::Bandwidth;
 use crate::recovery::rtt::RttStats;
 use crate::recovery::CongestionControlAlgorithm;
 use crate::recovery::HandshakeStatus;
 use crate::recovery::LossDetectionTimer;
+use crate::recovery::OnAckReceivedOutcome;
 use crate::recovery::RangeSet;
 use crate::recovery::RecoveryConfig;
 use crate::recovery::RecoveryOps;
+use crate::recovery::RecoveryStats;
 use crate::recovery::ReleaseDecision;
 use crate::recovery::Sent;
+use crate::recovery::StartupExit;
 use crate::recovery::GRANULARITY;
 use crate::recovery::INITIAL_PACKET_THRESHOLD;
 use crate::recovery::INITIAL_TIME_THRESHOLD;
 use crate::recovery::MAX_OUTSTANDING_NON_ACK_ELICITING;
 use crate::recovery::MAX_PACKET_THRESHOLD;
 use crate::recovery::MAX_PTO_PROBES_COUNT;
+use crate::Error;
+use crate::Result;
 
-use super::bandwidth::Bandwidth;
 use super::pacer::Pacer;
 use super::Acked;
 use super::Congestion;
@@ -53,7 +60,7 @@ enum SentStatus {
         ack_eliciting: bool,
         in_flight: bool,
         has_data: bool,
-        pmtud: bool,
+        is_pmtud_probe: bool,
         sent_bytes: usize,
         frames: SmallVec<[frame::Frame; 1]>,
     },
@@ -97,6 +104,10 @@ struct RecoveryEpoch {
 
     acked_frames: Vec<frame::Frame>,
     lost_frames: Vec<frame::Frame>,
+
+    /// The largest packet number sent in the packet number space so far.
+    #[allow(dead_code)]
+    test_largest_sent_pkt_num_on_path: Option<u64>,
 }
 
 struct AckedDetectionResult {
@@ -150,16 +161,11 @@ impl RecoveryEpoch {
         unacked_bytes
     }
 
+    // `peer_sent_ack_ranges` should not be used without validation.
     fn detect_and_remove_acked_packets(
-        &mut self, acked: &RangeSet, newly_acked: &mut Vec<Acked>, trace_id: &str,
-    ) -> AckedDetectionResult {
-        // Update the largest acked packet
-        self.largest_acked_packet.replace(
-            self.largest_acked_packet
-                .unwrap_or(0)
-                .max(acked.last().unwrap()),
-        );
-
+        &mut self, peer_sent_ack_ranges: &RangeSet, newly_acked: &mut Vec<Acked>,
+        skip_pn: Option<u64>, trace_id: &str,
+    ) -> Result<AckedDetectionResult> {
         newly_acked.clear();
 
         let mut acked_bytes = 0;
@@ -167,29 +173,41 @@ impl RecoveryEpoch {
         let mut spurious_pkt_thresh = None;
         let mut has_ack_eliciting = false;
 
-        let largest_acked = self.largest_acked_packet.unwrap();
+        let largest_ack_received = peer_sent_ack_ranges.last().unwrap();
+        let largest_acked = self
+            .largest_acked_packet
+            .unwrap_or(0)
+            .max(largest_ack_received);
+
+        for peer_sent_range in peer_sent_ack_ranges.iter() {
+            if skip_pn.is_some_and(|skip_pn| peer_sent_range.contains(&skip_pn)) {
+                // https://www.rfc-editor.org/rfc/rfc9000#section-13.1
+                // An endpoint SHOULD treat receipt of an acknowledgment
+                // for a packet it did not send as
+                // a connection error of type PROTOCOL_VIOLATION
+                return Err(Error::OptimisticAckDetected);
+            }
 
-        for ack in acked.iter() {
             // Because packets always have incrementing numbers, they are always
             // in sorted order.
             let start = if self
                 .sent_packets
                 .front()
-                .filter(|e| e.pkt_num >= ack.start)
+                .filter(|e| e.pkt_num >= peer_sent_range.start)
                 .is_some()
             {
                 // Usually it will be the first packet.
                 0
             } else {
                 self.sent_packets
-                    .binary_search_by_key(&ack.start, |p| p.pkt_num)
+                    .binary_search_by_key(&peer_sent_range.start, |p| p.pkt_num)
                     .unwrap_or_else(|e| e)
             };
 
             for SentPacket { pkt_num, status } in
                 self.sent_packets.range_mut(start..)
             {
-                if *pkt_num < ack.end {
+                if *pkt_num < peer_sent_range.end {
                     match status.ack() {
                         SentStatus::Sent {
                             time_sent,
@@ -231,12 +249,12 @@ impl RecoveryEpoch {
 
         self.drain_acked_and_lost_packets();
 
-        AckedDetectionResult {
+        Ok(AckedDetectionResult {
             acked_bytes,
             spurious_losses,
             spurious_pkt_thresh,
             has_ack_eliciting,
-        }
+        })
     }
 
     fn detect_and_remove_lost_packets(
@@ -265,7 +283,7 @@ impl RecoveryEpoch {
                         in_flight,
                         sent_bytes,
                         frames,
-                        pmtud,
+                        is_pmtud_probe,
                         ..
                     } = status.lose()
                     {
@@ -274,7 +292,7 @@ impl RecoveryEpoch {
                         if in_flight {
                             self.pkts_in_flight -= 1;
 
-                            if pmtud {
+                            if is_pmtud_probe {
                                 pmtud_lost_bytes += sent_bytes;
                                 pmtud_lost_packets.push(*pkt_num);
                                 // Do not track PMTUD probes losses
@@ -342,6 +360,8 @@ pub struct GRecovery {
 
     rtt_stats: RttStats,
 
+    recovery_stats: RecoveryStats,
+
     pub lost_count: usize,
 
     pub lost_spurious_count: usize,
@@ -350,7 +370,7 @@ pub struct GRecovery {
 
     time_thresh: f64,
 
-    bytes_in_flight: usize,
+    bytes_in_flight: BytesInFlight,
 
     bytes_sent: usize,
 
@@ -368,7 +388,7 @@ pub struct GRecovery {
     newly_acked: Vec<Acked>,
 
     /// A [`Vec`] that can be reused for calls of
-    /// [`detect_and_remove_lost_packets`] to avoid allocations
+    /// [`Self::detect_and_remove_lost_packets`] to avoid allocations
     lost_reuse: Vec<Lost>,
 
     pacer: Pacer,
@@ -388,6 +408,7 @@ impl GRecovery {
         Some(Self {
             epochs: Default::default(),
             rtt_stats: RttStats::new(recovery_config.max_ack_delay),
+            recovery_stats: RecoveryStats::default(),
             loss_timer: Default::default(),
             pto_count: 0,
 
@@ -397,7 +418,7 @@ impl GRecovery {
             pkt_thresh: INITIAL_PACKET_THRESHOLD,
             time_thresh: INITIAL_TIME_THRESHOLD,
 
-            bytes_in_flight: 0,
+            bytes_in_flight: Default::default(),
             bytes_sent: 0,
             bytes_lost: 0,
 
@@ -439,7 +460,8 @@ impl GRecovery {
             lost,
         );
 
-        self.bytes_in_flight -= lost_bytes + pmtud_lost_bytes;
+        self.bytes_in_flight
+            .saturating_subtract(lost_bytes + pmtud_lost_bytes, now);
 
         for pkt in pmtud_lost_packets {
             self.pacer.on_packet_neutered(pkt);
@@ -470,7 +492,7 @@ impl GRecovery {
         let mut duration = self.pto() * (1 << self.pto_count);
 
         // Arm PTO from now when there are no inflight packets.
-        if self.bytes_in_flight == 0 {
+        if self.bytes_in_flight.is_zero() {
             if handshake_status.has_handshake_keys {
                 return (Some(now + duration), packet::Epoch::Handshake);
             } else {
@@ -522,7 +544,9 @@ impl GRecovery {
             return;
         }
 
-        if self.bytes_in_flight == 0 && handshake_status.peer_verified_address {
+        if self.bytes_in_flight.is_zero() &&
+            handshake_status.peer_verified_address
+        {
             self.loss_timer.clear();
             return;
         }
@@ -590,8 +614,9 @@ impl RecoveryOps for GRecovery {
 
         let ack_eliciting = pkt.ack_eliciting;
         let in_flight = pkt.in_flight;
-        let sent_bytes = pkt.size;
+        let is_pmtud_probe = pkt.is_pmtud_probe;
         let pkt_num = pkt.pkt_num;
+        let sent_bytes = pkt.size;
 
         if let Some(SentPacket { pkt_num, .. }) = epoch.sent_packets.back() {
             assert!(*pkt_num < pkt.pkt_num, "Packet numbers must increase");
@@ -601,12 +626,19 @@ impl RecoveryOps for GRecovery {
             time_sent,
             ack_eliciting,
             in_flight,
-            pmtud: pkt.pmtud,
+            is_pmtud_probe,
             has_data: pkt.has_data,
             sent_bytes,
             frames: pkt.frames,
         };
 
+        #[cfg(test)]
+        {
+            epoch.test_largest_sent_pkt_num_on_path = epoch
+                .test_largest_sent_pkt_num_on_path
+                .max(Some(pkt.pkt_num));
+        }
+
         epoch.sent_packets.push_back(SentPacket { pkt_num, status });
 
         if ack_eliciting {
@@ -619,14 +651,14 @@ impl RecoveryOps for GRecovery {
         if in_flight {
             self.pacer.on_packet_sent(
                 time_sent,
-                self.bytes_in_flight,
+                self.bytes_in_flight.get(),
                 pkt_num,
                 sent_bytes,
                 pkt.has_data,
                 &self.rtt_stats,
             );
 
-            self.bytes_in_flight += sent_bytes;
+            self.bytes_in_flight.add(sent_bytes, now);
             epoch.pkts_in_flight += 1;
             self.set_loss_detection_timer(handshake_status, time_sent);
         }
@@ -636,16 +668,17 @@ impl RecoveryOps for GRecovery {
         trace!("{} {:?}", trace_id, self);
     }
 
-    fn get_packet_send_time(&self) -> Instant {
-        let now = Instant::now();
+    fn get_packet_send_time(&self, now: Instant) -> Instant {
         self.pacer.get_next_release_time().time(now).unwrap_or(now)
     }
 
+    // `peer_sent_ack_ranges` should not be used without validation.
     fn on_ack_received(
-        &mut self, ranges: &RangeSet, ack_delay: u64, epoch: packet::Epoch,
-        handshake_status: HandshakeStatus, now: Instant, trace_id: &str,
-    ) -> (usize, usize, usize) {
-        let prior_in_flight = self.bytes_in_flight;
+        &mut self, peer_sent_ack_ranges: &RangeSet, ack_delay: u64,
+        epoch: packet::Epoch, handshake_status: HandshakeStatus, now: Instant,
+        skip_pn: Option<u64>, trace_id: &str,
+    ) -> Result<OnAckReceivedOutcome> {
+        let prior_in_flight = self.bytes_in_flight.get();
 
         let AckedDetectionResult {
             acked_bytes,
@@ -653,10 +686,11 @@ impl RecoveryOps for GRecovery {
             spurious_pkt_thresh,
             has_ack_eliciting,
         } = self.epochs[epoch].detect_and_remove_acked_packets(
-            ranges,
+            peer_sent_ack_ranges,
             &mut self.newly_acked,
+            skip_pn,
             trace_id,
-        );
+        )?;
 
         self.lost_spurious_count += spurious_losses;
         if let Some(thresh) = spurious_pkt_thresh {
@@ -665,15 +699,23 @@ impl RecoveryOps for GRecovery {
         }
 
         if self.newly_acked.is_empty() {
-            return (0, 0, 0);
+            return Ok(OnAckReceivedOutcome::default());
         }
 
-        self.bytes_in_flight -= acked_bytes;
+        self.bytes_in_flight.saturating_subtract(acked_bytes, now);
+
+        let largest_newly_acked = self.newly_acked.last().unwrap();
+
+        // Update `largest_acked_packet` based on the validated `newly_acked`
+        // value.
+        let largest_acked_pkt_num = self.epochs[epoch]
+            .largest_acked_packet
+            .unwrap_or(0)
+            .max(largest_newly_acked.pkt_num);
+        self.epochs[epoch].largest_acked_packet = Some(largest_acked_pkt_num);
 
         // Check if largest packet is newly acked.
-        let largest_newly_acked = self.newly_acked.last().unwrap();
-        let update_rtt: bool = largest_newly_acked.pkt_num ==
-            ranges.last().unwrap() &&
+        let update_rtt = largest_newly_acked.pkt_num == largest_acked_pkt_num &&
             has_ack_eliciting;
         if update_rtt {
             let latest_rtt = now - largest_newly_acked.time_sent;
@@ -691,12 +733,13 @@ impl RecoveryOps for GRecovery {
         self.pacer.on_congestion_event(
             update_rtt,
             prior_in_flight,
-            self.bytes_in_flight,
+            self.bytes_in_flight.get(),
             now,
             &self.newly_acked,
             &self.lost_reuse,
             self.epochs[epoch].least_unacked(),
             &self.rtt_stats,
+            &mut self.recovery_stats,
         );
 
         self.pto_count = 0;
@@ -706,17 +749,22 @@ impl RecoveryOps for GRecovery {
 
         trace!("{} {:?}", trace_id, self);
 
-        (lost_packets, lost_bytes, acked_bytes)
+        Ok(OnAckReceivedOutcome {
+            lost_packets,
+            lost_bytes,
+            acked_bytes,
+            spurious_losses,
+        })
     }
 
     fn on_loss_detection_timeout(
         &mut self, handshake_status: HandshakeStatus, now: Instant,
         trace_id: &str,
-    ) -> (usize, usize) {
+    ) -> OnLossDetectionTimeoutOutcome {
         let (earliest_loss_time, epoch) = self.loss_time_and_space();
 
         if earliest_loss_time.is_some() {
-            let prior_in_flight = self.bytes_in_flight;
+            let prior_in_flight = self.bytes_in_flight.get();
 
             let (lost_bytes, lost_packets) =
                 self.detect_and_remove_lost_packets(epoch, now);
@@ -724,12 +772,13 @@ impl RecoveryOps for GRecovery {
             self.pacer.on_congestion_event(
                 false,
                 prior_in_flight,
-                self.bytes_in_flight,
+                self.bytes_in_flight.get(),
                 now,
                 &[],
                 &self.lost_reuse,
                 self.epochs[epoch].least_unacked(),
                 &self.rtt_stats,
+                &mut self.recovery_stats,
             );
 
             self.lost_count += lost_packets;
@@ -737,10 +786,13 @@ impl RecoveryOps for GRecovery {
             self.set_loss_detection_timer(handshake_status, now);
 
             trace!("{} {:?}", trace_id, self);
-            return (lost_packets, lost_bytes);
+            return OnLossDetectionTimeoutOutcome {
+                lost_packets,
+                lost_bytes,
+            };
         }
 
-        let epoch = if self.bytes_in_flight > 0 {
+        let epoch = if self.bytes_in_flight.get() > 0 {
             // Send new data if available, else retransmit old data. If neither
             // is available, send a single PING frame.
             let (_, e) = self.pto_time_and_space(handshake_status, now);
@@ -800,7 +852,10 @@ impl RecoveryOps for GRecovery {
         self.set_loss_detection_timer(handshake_status, now);
 
         trace!("{} {:?}", trace_id, self);
-        (0, 0)
+        OnLossDetectionTimeoutOutcome {
+            lost_packets: 0,
+            lost_bytes: 0,
+        }
     }
 
     fn on_pkt_num_space_discarded(
@@ -808,9 +863,8 @@ impl RecoveryOps for GRecovery {
         now: Instant,
     ) {
         let epoch = &mut self.epochs[epoch];
-        self.bytes_in_flight = self
-            .bytes_in_flight
-            .saturating_sub(epoch.discard(&mut self.pacer));
+        self.bytes_in_flight
+            .saturating_subtract(epoch.discard(&mut self.pacer), now);
         self.set_loss_detection_timer(handshake_status, now);
     }
 
@@ -837,7 +891,7 @@ impl RecoveryOps for GRecovery {
             return usize::MAX;
         }
 
-        self.cwnd().saturating_sub(self.bytes_in_flight)
+        self.cwnd().saturating_sub(self.bytes_in_flight.get())
     }
 
     fn rtt(&self) -> Duration {
@@ -848,6 +902,10 @@ impl RecoveryOps for GRecovery {
         self.rtt_stats.min_rtt()
     }
 
+    fn max_rtt(&self) -> Option<Duration> {
+        self.rtt_stats.max_rtt()
+    }
+
     fn rttvar(&self) -> Duration {
         self.rtt_stats.rttvar()
     }
@@ -857,10 +915,14 @@ impl RecoveryOps for GRecovery {
         r.rtt() + (r.rttvar() * 4).max(GRANULARITY)
     }
 
-    fn delivery_rate(&self) -> u64 {
-        self.pacer
-            .bandwidth_estimate(&self.rtt_stats)
-            .to_bits_per_second()
+    /// The most recent data delivery rate estimate.
+    fn delivery_rate(&self) -> Bandwidth {
+        self.pacer.bandwidth_estimate(&self.rtt_stats)
+    }
+
+    /// Statistics from when a CCA first exited the startup phase.
+    fn startup_exit(&self) -> Option<StartupExit> {
+        self.recovery_stats.startup_exit
     }
 
     fn max_datagram_size(&self) -> usize {
@@ -878,8 +940,9 @@ impl RecoveryOps for GRecovery {
         )
     }
 
+    // FIXME only used by gcongestion
     fn on_app_limited(&mut self) {
-        self.pacer.on_app_limited(self.bytes_in_flight)
+        self.pacer.on_app_limited(self.bytes_in_flight.get())
     }
 
     #[cfg(test)]
@@ -894,13 +957,17 @@ impl RecoveryOps for GRecovery {
 
     #[cfg(test)]
     fn bytes_in_flight(&self) -> usize {
-        self.bytes_in_flight
+        self.bytes_in_flight.get()
+    }
+
+    fn bytes_in_flight_duration(&self) -> Duration {
+        self.bytes_in_flight.get_duration()
     }
 
     #[cfg(test)]
     fn pacing_rate(&self) -> u64 {
         self.pacer
-            .pacing_rate(self.bytes_in_flight, &self.rtt_stats)
+            .pacing_rate(self.bytes_in_flight.get(), &self.rtt_stats)
             .to_bytes_per_period(Duration::from_secs(1))
     }
 
@@ -929,14 +996,21 @@ impl RecoveryOps for GRecovery {
     }
 
     #[cfg(test)]
-    fn app_limited(&self) -> bool {
-        self.pacer.is_app_limited(self.bytes_in_flight)
+    fn largest_sent_pkt_num_on_path(&self, epoch: packet::Epoch) -> Option<u64> {
+        self.epochs[epoch].test_largest_sent_pkt_num_on_path
     }
 
+    #[cfg(test)]
+    fn app_limited(&self) -> bool {
+        self.pacer.is_app_limited(self.bytes_in_flight.get())
+    }
+
+    // FIXME only used by congestion
     fn update_app_limited(&mut self, _v: bool) {
         // TODO
     }
 
+    // FIXME only used by congestion
     fn delivery_rate_update_app_limited(&mut self, _v: bool) {
         // TODO
     }
@@ -956,14 +1030,14 @@ impl RecoveryOps for GRecovery {
     #[cfg(feature = "qlog")]
     fn maybe_qlog(&mut self) -> Option<EventData> {
         let qlog_metrics = QlogMetrics {
-            min_rtt: self.rtt_stats.min_rtt().unwrap_or(Duration::MAX),
+            min_rtt: *self.rtt_stats.min_rtt,
             smoothed_rtt: self.rtt(),
             latest_rtt: self.rtt_stats.latest_rtt(),
             rttvar: self.rtt_stats.rttvar(),
             cwnd: self.cwnd() as u64,
-            bytes_in_flight: self.bytes_in_flight as u64,
+            bytes_in_flight: self.bytes_in_flight.get() as u64,
             ssthresh: self.pacer.ssthresh(),
-            pacing_rate: self.delivery_rate(),
+            pacing_rate: self.delivery_rate().to_bytes_per_second(),
         };
 
         self.qlog_metrics.maybe_update(qlog_metrics)
@@ -972,7 +1046,7 @@ impl RecoveryOps for GRecovery {
     fn send_quantum(&self) -> usize {
         let pacing_rate = self
             .pacer
-            .pacing_rate(self.bytes_in_flight, &self.rtt_stats);
+            .pacing_rate(self.bytes_in_flight.get(), &self.rtt_stats);
 
         let floor = if pacing_rate < Bandwidth::from_kbits_per_second(1200) {
             self.max_datagram_size
@@ -991,7 +1065,7 @@ impl std::fmt::Debug for GRecovery {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "timer={:?} ", self.loss_detection_timer())?;
         write!(f, "rtt_stats={:?} ", self.rtt_stats)?;
-        write!(f, "bytes_in_flight={} ", self.bytes_in_flight)?;
+        write!(f, "bytes_in_flight={} ", self.bytes_in_flight.get())?;
         write!(f, "{:?} ", self.pacer)?;
         Ok(())
     }
