@@ -31,7 +31,9 @@ use std::time::Instant;
 use crate::frame;
 use crate::packet;
 use crate::ranges::RangeSet;
+use crate::recovery::bandwidth::Bandwidth;
 use crate::Config;
+use crate::Result;
 
 #[cfg(feature = "qlog")]
 use qlog::events::EventData;
@@ -130,6 +132,20 @@ pub(crate) enum Recovery {
     GCongestion(GRecovery),
 }
 
+#[derive(Debug, Default, PartialEq)]
+pub struct OnAckReceivedOutcome {
+    pub lost_packets: usize,
+    pub lost_bytes: usize,
+    pub acked_bytes: usize,
+    pub spurious_losses: usize,
+}
+
+#[derive(Debug, Default)]
+pub struct OnLossDetectionTimeoutOutcome {
+    pub lost_packets: usize,
+    pub lost_bytes: usize,
+}
+
 #[enum_dispatch::enum_dispatch]
 /// Api for the Recovery implementation
 pub trait RecoveryOps {
@@ -156,17 +172,18 @@ pub trait RecoveryOps {
         &mut self, pkt: Sent, epoch: packet::Epoch,
         handshake_status: HandshakeStatus, now: Instant, trace_id: &str,
     );
-    fn get_packet_send_time(&self) -> Instant;
+    fn get_packet_send_time(&self, now: Instant) -> Instant;
 
     fn on_ack_received(
         &mut self, ranges: &RangeSet, ack_delay: u64, epoch: packet::Epoch,
-        handshake_status: HandshakeStatus, now: Instant, trace_id: &str,
-    ) -> (usize, usize, usize);
+        handshake_status: HandshakeStatus, now: Instant, skip_pn: Option<u64>,
+        trace_id: &str,
+    ) -> Result<OnAckReceivedOutcome>;
 
     fn on_loss_detection_timeout(
         &mut self, handshake_status: HandshakeStatus, now: Instant,
         trace_id: &str,
-    ) -> (usize, usize);
+    ) -> OnLossDetectionTimeoutOutcome;
     fn on_pkt_num_space_discarded(
         &mut self, epoch: packet::Epoch, handshake_status: HandshakeStatus,
         now: Instant,
@@ -181,11 +198,17 @@ pub trait RecoveryOps {
 
     fn min_rtt(&self) -> Option<Duration>;
 
+    fn max_rtt(&self) -> Option<Duration>;
+
     fn rttvar(&self) -> Duration;
 
     fn pto(&self) -> Duration;
 
-    fn delivery_rate(&self) -> u64;
+    /// The most recent data delivery rate estimate.
+    fn delivery_rate(&self) -> Bandwidth;
+
+    /// Statistics from when a CCA first exited the startup phase.
+    fn startup_exit(&self) -> Option<StartupExit>;
 
     fn max_datagram_size(&self) -> usize;
 
@@ -195,6 +218,11 @@ pub trait RecoveryOps {
 
     fn on_app_limited(&mut self);
 
+    // Since a recovery module is path specific, this tracks the largest packet
+    // sent per path.
+    #[cfg(test)]
+    fn largest_sent_pkt_num_on_path(&self, epoch: packet::Epoch) -> Option<u64>;
+
     #[cfg(test)]
     fn app_limited(&self) -> bool;
 
@@ -204,6 +232,8 @@ pub trait RecoveryOps {
     #[cfg(test)]
     fn bytes_in_flight(&self) -> usize;
 
+    fn bytes_in_flight_duration(&self) -> Duration;
+
     #[cfg(test)]
     fn in_flight_count(&self, epoch: packet::Epoch) -> usize;
 
@@ -328,7 +358,7 @@ pub struct Sent {
 
     pub has_data: bool,
 
-    pub pmtud: bool,
+    pub is_pmtud_probe: bool,
 }
 
 impl std::fmt::Debug for Sent {
@@ -343,7 +373,7 @@ impl std::fmt::Debug for Sent {
         write!(f, "tx_in_flight={} ", self.tx_in_flight)?;
         write!(f, "lost={} ", self.lost)?;
         write!(f, "has_data={} ", self.has_data)?;
-        write!(f, "pmtud={}", self.pmtud)?;
+        write!(f, "is_pmtud_probe={}", self.is_pmtud_probe)?;
 
         Ok(())
     }
@@ -554,13 +584,59 @@ impl ReleaseDecision {
     }
 }
 
+/// Recovery statistics
+#[derive(Default, Debug)]
+pub struct RecoveryStats {
+    startup_exit: Option<StartupExit>,
+}
+
+impl RecoveryStats {
+    // Record statistics when a CCA first exits startup.
+    pub fn set_startup_exit(&mut self, startup_exit: StartupExit) {
+        if self.startup_exit.is_none() {
+            self.startup_exit = Some(startup_exit);
+        }
+    }
+}
+
+/// Statistics from when a CCA first exited the startup phase.
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub struct StartupExit {
+    /// The congestion_window recorded at Startup exit.
+    pub cwnd: usize,
+
+    /// The reason a CCA exited the startup phase.
+    pub reason: StartupExitReason,
+}
+
+impl StartupExit {
+    fn new(cwnd: usize, reason: StartupExitReason) -> Self {
+        Self { cwnd, reason }
+    }
+}
+
+/// The reason a CCA exited the startup phase.
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub enum StartupExitReason {
+    /// Exit startup due to excessive loss
+    Loss,
+
+    /// Exit startup due to bandwidth plateau.
+    BandwidthPlateau,
+
+    /// Exit startup due to persistent queue.
+    PersistentQueue,
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
     use crate::packet;
     use crate::ranges;
     use crate::recovery::congestion::PACING_MULTIPLIER;
+    use crate::testing;
     use crate::CongestionControlAlgorithm;
+    use rstest::rstest;
     use smallvec::smallvec;
     use std::str::FromStr;
 
@@ -610,10 +686,13 @@ mod tests {
         );
     }
 
-    #[test]
-    fn loss_on_pto() {
+    #[rstest]
+    fn loss_on_pto(
+        #[values("reno", "cubic", "bbr", "bbr2", "bbr2_gcongestion")]
+        cc_algorithm_name: &str,
+    ) {
         let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
-        cfg.set_cc_algorithm(CongestionControlAlgorithm::Reno);
+        assert_eq!(cfg.set_cc_algorithm_name(cc_algorithm_name), Ok(()));
 
         let mut r = Recovery::new(&cfg);
 
@@ -638,7 +717,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -651,6 +730,7 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 1);
         assert_eq!(r.bytes_in_flight(), 1000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 1,
@@ -668,7 +748,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -681,6 +761,7 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 2000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 2,
@@ -698,7 +779,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -710,6 +791,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 3);
         assert_eq!(r.bytes_in_flight(), 3000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 3,
@@ -727,7 +809,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -739,6 +821,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 4);
         assert_eq!(r.bytes_in_flight(), 4000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         // Wait for 10ms.
         now += Duration::from_millis(10);
@@ -754,13 +837,21 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, 2 * 1000)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 2 * 1000,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 2000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(10));
         assert_eq!(r.lost_count(), 0);
 
         // Wait until loss detection timer expires.
@@ -788,7 +879,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -800,6 +891,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 3);
         assert_eq!(r.bytes_in_flight(), 3000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(30));
 
         let p = Sent {
             pkt_num: 5,
@@ -817,7 +909,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -829,6 +921,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 4);
         assert_eq!(r.bytes_in_flight(), 4000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(30));
         assert_eq!(r.lost_count(), 0);
 
         // Wait for 10ms.
@@ -845,13 +938,21 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (2, 2000, 2 * 1000)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 2,
+                lost_bytes: 2000,
+                acked_bytes: 2 * 1000,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 4);
         assert_eq!(r.bytes_in_flight(), 0);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(40));
 
         assert_eq!(r.lost_count(), 2);
 
@@ -864,12 +965,21 @@ mod tests {
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
+        if cc_algorithm_name == "reno" || cc_algorithm_name == "cubic" {
+            assert!(r.startup_exit().is_some());
+            assert_eq!(r.startup_exit().unwrap().reason, StartupExitReason::Loss);
+        } else {
+            assert_eq!(r.startup_exit(), None);
+        }
     }
 
-    #[test]
-    fn loss_on_timer() {
+    #[rstest]
+    fn loss_on_timer(
+        #[values("reno", "cubic", "bbr", "bbr2", "bbr2_gcongestion")]
+        cc_algorithm_name: &str,
+    ) {
         let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
-        cfg.set_cc_algorithm(CongestionControlAlgorithm::Reno);
+        assert_eq!(cfg.set_cc_algorithm_name(cc_algorithm_name), Ok(()));
 
         let mut r = Recovery::new(&cfg);
 
@@ -894,7 +1004,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -906,6 +1016,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 1);
         assert_eq!(r.bytes_in_flight(), 1000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 1,
@@ -923,7 +1034,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -935,6 +1046,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 2000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 2,
@@ -952,7 +1064,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -964,6 +1076,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 3);
         assert_eq!(r.bytes_in_flight(), 3000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 3,
@@ -981,7 +1094,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -993,6 +1106,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 4);
         assert_eq!(r.bytes_in_flight(), 4000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         // Wait for 10ms.
         now += Duration::from_millis(10);
@@ -1009,13 +1123,21 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, 3 * 1000)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 3 * 1000,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 1000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(10));
         assert_eq!(r.lost_count(), 0);
 
         // Wait until loss detection timer expires.
@@ -1027,6 +1149,7 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 0);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_micros(11250));
 
         assert_eq!(r.lost_count(), 1);
 
@@ -1039,12 +1162,21 @@ mod tests {
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
+        if cc_algorithm_name == "reno" || cc_algorithm_name == "cubic" {
+            assert!(r.startup_exit().is_some());
+            assert_eq!(r.startup_exit().unwrap().reason, StartupExitReason::Loss);
+        } else {
+            assert_eq!(r.startup_exit(), None);
+        }
     }
 
-    #[test]
-    fn loss_on_reordering() {
+    #[rstest]
+    fn loss_on_reordering(
+        #[values("reno", "cubic", "bbr", "bbr2", "bbr2_gcongestion")]
+        cc_algorithm_name: &str,
+    ) {
         let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
-        cfg.set_cc_algorithm(CongestionControlAlgorithm::Reno);
+        assert_eq!(cfg.set_cc_algorithm_name(cc_algorithm_name), Ok(()));
 
         let mut r = Recovery::new(&cfg);
 
@@ -1069,7 +1201,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1081,6 +1213,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 1);
         assert_eq!(r.bytes_in_flight(), 1000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 1,
@@ -1098,7 +1231,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1110,6 +1243,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 2000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 2,
@@ -1127,7 +1261,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1139,6 +1273,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 3);
         assert_eq!(r.bytes_in_flight(), 3000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 3,
@@ -1156,7 +1291,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1168,6 +1303,7 @@ mod tests {
         );
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 4);
         assert_eq!(r.bytes_in_flight(), 4000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         // Wait for 10ms.
         now += Duration::from_millis(10);
@@ -1183,9 +1319,16 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (1, 1000, 1000 * 2)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 1,
+                lost_bytes: 1000,
+                acked_bytes: 1000 * 2,
+                spurious_losses: 0,
+            }
         );
 
         now += Duration::from_millis(10);
@@ -1202,13 +1345,21 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, 1000)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 1000,
+                spurious_losses: 1,
+            }
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
         assert_eq!(r.bytes_in_flight(), 0);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(20));
 
         // Spurious loss.
         assert_eq!(r.lost_count(), 1);
@@ -1226,12 +1377,21 @@ mod tests {
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
+        if cc_algorithm_name == "reno" || cc_algorithm_name == "cubic" {
+            assert!(r.startup_exit().is_some());
+            assert_eq!(r.startup_exit().unwrap().reason, StartupExitReason::Loss);
+        } else {
+            assert_eq!(r.startup_exit(), None);
+        }
     }
 
-    #[test]
-    fn pacing() {
+    #[rstest]
+    fn pacing(
+        #[values("reno", "cubic", "bbr", "bbr2", "bbr2_gcongestion")]
+        cc_algorithm_name: &str,
+    ) {
         let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
-        cfg.set_cc_algorithm(CongestionControlAlgorithm::CUBIC);
+        assert_eq!(cfg.set_cc_algorithm_name(cc_algorithm_name), Ok(()));
 
         let mut r = Recovery::new(&cfg);
 
@@ -1239,46 +1399,56 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
 
-        // send out first packet (a full initcwnd).
-        let p = Sent {
-            pkt_num: 0,
-            frames: smallvec![],
-            time_sent: now,
-            time_acked: None,
-            time_lost: None,
-            size: 12000,
-            ack_eliciting: true,
-            in_flight: true,
-            delivered: 0,
-            delivered_time: now,
-            first_sent_time: now,
-            is_app_limited: false,
-            tx_in_flight: 0,
-            lost: 0,
-            has_data: false,
-            pmtud: false,
-        };
+        // send out first packet burst (a full initcwnd).
+        for i in 0..10 {
+            let p = Sent {
+                pkt_num: i,
+                frames: smallvec![],
+                time_sent: now,
+                time_acked: None,
+                time_lost: None,
+                size: 1200,
+                ack_eliciting: true,
+                in_flight: true,
+                delivered: 0,
+                delivered_time: now,
+                first_sent_time: now,
+                is_app_limited: false,
+                tx_in_flight: 0,
+                lost: 0,
+                has_data: true,
+                is_pmtud_probe: false,
+            };
 
-        r.on_packet_sent(
-            p,
-            packet::Epoch::Application,
-            HandshakeStatus::default(),
-            now,
-            "",
-        );
+            r.on_packet_sent(
+                p,
+                packet::Epoch::Application,
+                HandshakeStatus::default(),
+                now,
+                "",
+            );
+        }
 
-        assert_eq!(r.sent_packets_len(packet::Epoch::Application), 1);
+        assert_eq!(r.sent_packets_len(packet::Epoch::Application), 10);
         assert_eq!(r.bytes_in_flight(), 12000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
-        // First packet will be sent out immediately.
-        assert_eq!(r.pacing_rate(), 0);
-        assert_eq!(r.get_packet_send_time(), now);
+        // Next packet will be sent out immediately.
+        if cc_algorithm_name != "bbr2_gcongestion" {
+            assert_eq!(r.pacing_rate(), 0);
+        } else {
+            assert_eq!(r.pacing_rate(), 103963);
+        }
+        assert_eq!(r.get_packet_send_time(now), now);
+
+        assert_eq!(r.cwnd(), 12000);
+        assert_eq!(r.cwnd_available(), 0);
 
         // Wait 50ms for ACK.
         now += Duration::from_millis(50);
 
         let mut acked = ranges::RangeSet::default();
-        acked.insert(0..1);
+        acked.insert(0..10);
 
         assert_eq!(
             r.on_ack_received(
@@ -1287,21 +1457,29 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, 12000)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 12000,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
         assert_eq!(r.bytes_in_flight(), 0);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(50));
         assert_eq!(r.rtt(), Duration::from_millis(50));
 
-        // 1 MSS increased.
-        assert_eq!(r.cwnd(), 12000 + 1200);
+        // 10 MSS increased due to acks.
+        assert_eq!(r.cwnd(), 12000 + 1200 * 10);
 
-        // Send out second packet.
+        // Send the second packet burst.
         let p = Sent {
-            pkt_num: 1,
+            pkt_num: 10,
             frames: smallvec![],
             time_sent: now,
             time_acked: None,
@@ -1315,8 +1493,8 @@ mod tests {
             is_app_limited: false,
             tx_in_flight: 0,
             lost: 0,
-            has_data: false,
-            pmtud: false,
+            has_data: true,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1329,13 +1507,19 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 1);
         assert_eq!(r.bytes_in_flight(), 6000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(50));
 
-        // Pacing is not done during initial phase of connection.
-        assert_eq!(r.get_packet_send_time(), now);
+        if cc_algorithm_name != "bbr2_gcongestion" {
+            // Pacing is not done during initial phase of connection.
+            assert_eq!(r.get_packet_send_time(now), now);
+        } else {
+            // Pacing is done from the beginning.
+            assert_ne!(r.get_packet_send_time(now), now);
+        }
 
-        // Send the third packet out.
+        // Send the third packet burst.
         let p = Sent {
-            pkt_num: 2,
+            pkt_num: 11,
             frames: smallvec![],
             time_sent: now,
             time_acked: None,
@@ -1349,8 +1533,8 @@ mod tests {
             is_app_limited: false,
             tx_in_flight: 0,
             lost: 0,
-            has_data: false,
-            pmtud: false,
+            has_data: true,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1363,10 +1547,11 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 12000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(50));
 
-        // Send the third packet out.
+        // Send the fourth packet burst.
         let p = Sent {
-            pkt_num: 3,
+            pkt_num: 12,
             frames: smallvec![],
             time_sent: now,
             time_acked: None,
@@ -1380,8 +1565,8 @@ mod tests {
             is_app_limited: false,
             tx_in_flight: 0,
             lost: 0,
-            has_data: false,
-            pmtud: false,
+            has_data: true,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1394,24 +1579,159 @@ mod tests {
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 3);
         assert_eq!(r.bytes_in_flight(), 13000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(50));
 
         // We pace this outgoing packet. as all conditions for pacing
         // are passed.
-        let pacing_rate = (r.cwnd() as f64 * PACING_MULTIPLIER / 0.05) as u64;
+        let pacing_rate = match cc_algorithm_name {
+            "bbr" => {
+                // Constants from congestion/bbr/mod.rs
+                let cwnd_gain = 2.0;
+                let startup_pacing_gain = 2.89;
+                // Adjust for cwnd_gain.  BW estimate was made before the CWND
+                // increase.
+                let bw = r.cwnd() as f64 /
+                    cwnd_gain /
+                    Duration::from_millis(50).as_secs_f64();
+                (bw * startup_pacing_gain) as u64
+            },
+            "bbr2_gcongestion" => {
+                let cwnd_gain: f64 = 2.0;
+                // Adjust for cwnd_gain.  BW estimate was made before the CWND
+                // increase.
+                let bw = r.cwnd() as f64 /
+                    cwnd_gain /
+                    Duration::from_millis(50).as_secs_f64();
+                bw as u64
+            },
+            "bbr2" => {
+                // Constants from congestion/bbr2/mod.rs
+                let cwnd_gain = 2.0;
+                let startup_pacing_gain = 2.77;
+                let pacing_margin_percent = 0.01;
+                // Adjust for cwnd_gain.  BW estimate was made before the CWND
+                // increase.
+                let bw = r.cwnd() as f64 /
+                    cwnd_gain /
+                    Duration::from_millis(50).as_secs_f64();
+                (bw * startup_pacing_gain * (1.0 - pacing_margin_percent)) as u64
+            },
+            _ => {
+                let bw =
+                    r.cwnd() as f64 / Duration::from_millis(50).as_secs_f64();
+                (bw * PACING_MULTIPLIER) as u64
+            },
+        };
         assert_eq!(r.pacing_rate(), pacing_rate);
 
+        let scale_factor = if cc_algorithm_name == "bbr2_gcongestion" {
+            // For bbr2_gcongestion, send time is almost 13000 / pacing_rate.
+            // Don't know where 13000 comes from.
+            1.08333332
+        } else {
+            1.0
+        };
         assert_eq!(
-            r.get_packet_send_time(),
-            now + Duration::from_secs_f64(12000.0 / pacing_rate as f64)
+            r.get_packet_send_time(now) - now,
+            Duration::from_secs_f64(scale_factor * 12000.0 / pacing_rate as f64)
         );
+        assert_eq!(r.startup_exit(), None);
     }
 
-    #[test]
-    fn pmtud_loss_on_timer() {
+    #[rstest]
+    fn validate_ack_range_on_ack_received(
+        #[values("cubic", "bbr2", "bbr2_gcongestion")] cc_algorithm_name: &str,
+    ) {
         let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
-        cfg.set_cc_algorithm(CongestionControlAlgorithm::Reno);
+        cfg.set_cc_algorithm_name(cc_algorithm_name).unwrap();
+
+        let epoch = packet::Epoch::Application;
+        let mut r = Recovery::new(&cfg);
+        let mut now = Instant::now();
+        assert_eq!(r.sent_packets_len(epoch), 0);
+
+        // Send 4 packets
+        let pkt_size = 1000;
+        let pkt_count = 4;
+        for pkt_num in 0..pkt_count {
+            let sent = crate::testing::helper_packet_sent(pkt_num, now, pkt_size);
+            r.on_packet_sent(sent, epoch, HandshakeStatus::default(), now, "");
+        }
+        assert_eq!(r.sent_packets_len(epoch), pkt_count as usize);
+        assert_eq!(r.bytes_in_flight(), pkt_count as usize * pkt_size);
+        assert!(r.get_largest_acked_on_epoch(epoch).is_none());
+        assert_eq!(r.largest_sent_pkt_num_on_path(epoch).unwrap(), 3);
+
+        // Wait for 10ms.
+        now += Duration::from_millis(10);
+
+        // ACK 2 packets
+        let mut acked = ranges::RangeSet::default();
+        acked.insert(0..2);
+
+        assert_eq!(
+            r.on_ack_received(
+                &acked,
+                25,
+                epoch,
+                HandshakeStatus::default(),
+                now,
+                None,
+                "",
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 2 * 1000,
+                spurious_losses: 0,
+            }
+        );
+
+        assert_eq!(r.sent_packets_len(epoch), 2);
+        assert_eq!(r.bytes_in_flight(), 2 * 1000);
+
+        assert_eq!(r.get_largest_acked_on_epoch(epoch).unwrap(), 1);
+        assert_eq!(r.largest_sent_pkt_num_on_path(epoch).unwrap(), 3);
+
+        // ACK large range
+        let mut acked = ranges::RangeSet::default();
+        acked.insert(0..10);
+        assert_eq!(
+            r.on_ack_received(
+                &acked,
+                25,
+                epoch,
+                HandshakeStatus::default(),
+                now,
+                None,
+                "",
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 2 * 1000,
+                spurious_losses: 0,
+            }
+        );
+        assert_eq!(r.sent_packets_len(epoch), 0);
+        assert_eq!(r.bytes_in_flight(), 0);
+
+        assert_eq!(r.get_largest_acked_on_epoch(epoch).unwrap(), 3);
+        assert_eq!(r.largest_sent_pkt_num_on_path(epoch).unwrap(), 3);
+    }
+
+    #[rstest]
+    fn pmtud_loss_on_timer(
+        #[values("reno", "cubic", "bbr", "bbr2", "bbr2_gcongestion")]
+        cc_algorithm_name: &str,
+    ) {
+        let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
+        assert_eq!(cfg.set_cc_algorithm_name(cc_algorithm_name), Ok(()));
 
         let mut r = Recovery::new(&cfg);
+        assert_eq!(r.cwnd(), 12000);
 
         let mut now = Instant::now();
 
@@ -1434,7 +1754,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1448,6 +1768,7 @@ mod tests {
         assert_eq!(r.in_flight_count(packet::Epoch::Application), 1);
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 1);
         assert_eq!(r.bytes_in_flight(), 1000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::ZERO);
 
         let p = Sent {
             pkt_num: 1,
@@ -1465,7 +1786,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: true,
+            is_pmtud_probe: true,
         };
 
         r.on_packet_sent(
@@ -1494,7 +1815,7 @@ mod tests {
             tx_in_flight: 0,
             lost: 0,
             has_data: false,
-            pmtud: false,
+            is_pmtud_probe: false,
         };
 
         r.on_packet_sent(
@@ -1522,13 +1843,21 @@ mod tests {
                 packet::Epoch::Application,
                 HandshakeStatus::default(),
                 now,
+                None,
                 "",
-            ),
-            (0, 0, 2 * 1000)
+            )
+            .unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: 2 * 1000,
+                spurious_losses: 0,
+            }
         );
 
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.bytes_in_flight(), 1000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_millis(10));
         assert_eq!(r.lost_count(), 0);
 
         // Wait until loss detection timer expires.
@@ -1541,7 +1870,12 @@ mod tests {
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 2);
         assert_eq!(r.in_flight_count(packet::Epoch::Application), 0);
         assert_eq!(r.bytes_in_flight(), 0);
-        assert_eq!(r.cwnd(), 12000);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_micros(11250));
+        assert_eq!(r.cwnd(), match cc_algorithm_name {
+            "bbr" => 14000,
+            "bbr2" => 14000,
+            _ => 12000,
+        });
 
         assert_eq!(r.lost_count(), 0);
 
@@ -1556,10 +1890,75 @@ mod tests {
         assert_eq!(r.sent_packets_len(packet::Epoch::Application), 0);
         assert_eq!(r.in_flight_count(packet::Epoch::Application), 0);
         assert_eq!(r.bytes_in_flight(), 0);
+        assert_eq!(r.bytes_in_flight_duration(), Duration::from_micros(11250));
         assert_eq!(r.lost_count(), 0);
+        assert_eq!(r.startup_exit(), None);
+    }
+
+    // Modeling delivery_rate for gcongestion is non-trivial so we only test the
+    // congestion specific algorithms.
+    #[rstest]
+    fn congestion_delivery_rate(
+        #[values("reno", "cubic", "bbr", "bbr2")] cc_algorithm_name: &str,
+    ) {
+        let mut cfg = crate::Config::new(crate::PROTOCOL_VERSION).unwrap();
+        assert_eq!(cfg.set_cc_algorithm_name(cc_algorithm_name), Ok(()));
+
+        let mut r = Recovery::new(&cfg);
+        assert_eq!(r.cwnd(), 12000);
+
+        let now = Instant::now();
+
+        let mut total_bytes_sent = 0;
+        for pn in 0..10 {
+            // Start by sending a few packets.
+            let bytes = 1000;
+            let sent = testing::helper_packet_sent(pn, now, bytes);
+            r.on_packet_sent(
+                sent,
+                packet::Epoch::Application,
+                HandshakeStatus::default(),
+                now,
+                "",
+            );
+
+            total_bytes_sent += bytes;
+        }
+
+        // Ack
+        let interval = Duration::from_secs(10);
+        let mut acked = ranges::RangeSet::default();
+        acked.insert(0..10);
+        assert_eq!(
+            r.on_ack_received(
+                &acked,
+                25,
+                packet::Epoch::Application,
+                HandshakeStatus::default(),
+                now + interval,
+                None,
+                "",
+            ).unwrap(),
+            OnAckReceivedOutcome {
+                lost_packets: 0,
+                lost_bytes: 0,
+                acked_bytes: total_bytes_sent,
+                spurious_losses: 0,
+            }
+        );
+        assert_eq!(r.delivery_rate().to_bytes_per_second(), 1000);
+        assert_eq!(r.min_rtt().unwrap(), interval);
+        // delivery rate should be in units bytes/sec
+        assert_eq!(
+            total_bytes_sent as u64 / interval.as_secs(),
+            r.delivery_rate().to_bytes_per_second()
+        );
+        assert_eq!(r.startup_exit(), None);
     }
 }
 
+mod bandwidth;
+mod bytes_in_flight;
 mod congestion;
 mod gcongestion;
 mod rtt;
